const std = @import("std");
const mem = std.mem;
const net = std.net;
const http = std.http;
const heap = std.heap;

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    const address = try net.Address.parseIp("0.0.0.0", 8080);
    var listener = Listener.init(.{
        .address = address,
        .allocator = allocator,
        .root_handler = &handleError,
    });
    defer listener.deinit();

    try listener.router.addRoute("/", &handle);
    try listener.router.addRoute("/error", &handleError);
    try listener.router.addRoute("//pee", &handleError);
    try listener.router.addRoute("/error/*", &handleError);

    try listener.listen();
}

pub fn handle(event: *Listener.Event) anyerror!void {
    try event.res.body.appendSlice("hello there");
}

pub fn handleError(event: *Listener.Event) anyerror!void {
    try event.res.body.appendSlice("ahoy, an error occurred");
}

pub const Listener = struct {
    address: net.Address,
    allocator: mem.Allocator,

    router: Router,

    pub const Options = struct {
        address: net.Address,
        allocator: mem.Allocator,
        root_handler: HandlerFn,
    };

    pub fn init(options: Options) Listener {
        return .{
            .address = options.address,
            .allocator = options.allocator,
            .router = Router.init(options.allocator, options.root_handler),
        };
    }

    pub fn deinit(self: *Listener) void {
        self.router.deinit();
    }

    pub fn listen(self: *Listener) !void {
        var arena = heap.ArenaAllocator.init(self.allocator);
        defer arena.deinit();

        var tcp = try self.address.listen(.{ .reuse_address = true });
        defer tcp.deinit();
        std.debug.print("listening at: {any}\n", .{self.address});

        while (true) {
            const read_buf = try arena.allocator().alloc(u8, 1024 * 32);
            const connection = try tcp.accept();
            var server = http.Server.init(connection, read_buf);
            try self.handle(&server, arena.allocator());
            _ = arena.reset(.retain_capacity);
        }
    }

    fn handle(self: *Listener, server: *http.Server, allocator: mem.Allocator) !void {
        handler: while (true) {
            var req = server.receiveHead() catch |e| if (e == error.HttpConnectionClosing) break :handler else return e;

            var event = Event{
                .req = .{
                    .uri = try std.Uri.parseWithoutScheme(req.head.target),
                    .headers = std.ArrayList(*const http.Header).init(allocator),
                },
                .res = .{
                    .status = .ok,
                    .headers = std.StringHashMap(*http.Header).init(allocator),
                    .body = std.ArrayList(u8).init(allocator),
                },
            };

            var header_it = req.iterateHeaders();
            while (header_it.next()) |header| try event.req.headers.append(&header);

            try self.router.handle(&event);

            try respondFromEvent(&event, &req, allocator);
        }
    }

    fn respondFromEvent(event: *Event, req: *http.Server.Request, allocator: mem.Allocator) !void {
        const res_body = try event.res.body.toOwnedSlice();
        const res_headers = try allocator.alloc(http.Header, event.res.headers.count());
        var i: usize = 0;
        var header_it = event.res.headers.iterator();
        while (header_it.next()) |header_ptr| : (i += 1) res_headers[i] = header_ptr.value_ptr.*.*;
        try req.respond(res_body, .{
            .status = event.res.status,
            .extra_headers = res_headers,
        });
    }

    pub const Event = struct {
        req: Request,
        res: Response,

        pub const Request = struct {
            uri: std.Uri,
            // head: http.Server.Request.Head,
            headers: std.ArrayList(*const http.Header),
        };
        pub const Response = struct {
            status: http.Status,
            headers: std.StringHashMap(*http.Header),
            body: std.ArrayList(u8),
        };
    };

    pub const HandlerFn = *const fn (event: *Event) anyerror!void;

    pub const Router = struct {
        allocator: mem.Allocator,

        root_node: Node,
        static_routes: std.StringHashMap(*Node),

        pub fn init(allocator: mem.Allocator, root_handler: HandlerFn) Router {
            return .{
                .allocator = allocator,
                .root_node = .{
                    .type = .normal,
                    .max_depth = 1024,
                    .children = std.StringHashMap(*Node).init(allocator),
                    .handler = root_handler,
                    .placeholder_children = std.ArrayList(*Node).init(allocator),
                },
                .static_routes = std.StringHashMap(*Node).init(allocator),
            };
        }

        pub fn deinit(self: *Router) void {
            self.root_node.children.deinit();
            self.root_node.placeholder_children.deinit();
            self.static_routes.deinit();
        }

        pub fn handle(self: *Router, event: *Listener.Event) !void {
            const handlerFn = try self.route(event.req.uri.path);
            try handlerFn(event);
        }

        pub fn addRoute(self: *Router, path: []const u8, handler: HandlerFn) !void {
            // std.debug.print("\npath {s}\n", .{path});
            var is_static_route = true;
            var sections = mem.splitScalar(u8, path, '/');
            var node = &self.root_node;
            var unnamed_placeholder_ctr: usize = 0;
            var matched_nodes = std.ArrayList(*Node).init(self.allocator);
            defer matched_nodes.deinit();
            try matched_nodes.append(node);
            while (sections.next()) |section| {
                // std.debug.print("adding section {s}\n", .{section});
                if (node.children.get(section)) |child| {
                    node = child;
                } else {
                    var child_node = try Node.init(self.allocator);
                    child_node.type = if (mem.startsWith(u8, section, "**"))
                        .wildcard
                    else if ((section.len > 0 and section[0] == ':') or mem.eql(u8, section, "*"))
                        .placeholder
                    else
                        .normal;
                    child_node.parent = node;
                    try node.children.put(section, child_node);

                    switch (child_node.type) {
                        .normal => {},
                        .wildcard => {
                            // std.debug.print("is wildcard\n", .{});
                            node.wildcard_child_node = child_node;
                            child_node.param_name = if (section.len > 3) section[3..] else "_";
                            is_static_route = false;
                        },
                        .placeholder => {
                            // std.debug.print("is placeholder\n", .{});
                            child_node.param_name = if (mem.eql(u8, section, "*")) blk: {
                                // std.debug.print("is unnamed placeholder #{d}\n", .{unnamed_placeholder_ctr});
                                const s = try std.fmt.allocPrint(self.allocator, "_{d}", .{unnamed_placeholder_ctr}); // TODO: this will leak
                                unnamed_placeholder_ctr += 1;
                                break :blk s;
                            } else section[1..];
                            try node.placeholder_children.append(child_node);
                            is_static_route = false;
                        },
                    }

                    try matched_nodes.append(child_node);
                    node = child_node;
                }
            }

            node.handler = handler;

            // if (is_static_route) std.debug.print("was static route\n", .{});

            if (is_static_route) try self.static_routes.put(path, node);
        }

        pub fn route(self: *Router, path: []const u8) !HandlerFn {
            if (self.static_routes.get(path)) |rt| return rt.handler;

            var params = std.StringHashMap([]const u8).init(self.allocator);
            defer params.deinit();
            var params_found = false;
            var wildcard_node: ?*Node = null;
            var node: *Node = &self.root_node;
            var wildcard_param: ?[]const u8 = null;

            var remaining = mem.count(u8, path, "/") + 1;
            var sections = mem.splitScalar(u8, path, '/');
            while (sections.next()) |section| : (remaining -= 1) {
                if (node.wildcard_child_node) |wildcard_child_node| {
                    wildcard_node = wildcard_child_node;
                    wildcard_param = sections.rest();
                }

                const next_node = node.children.get(section);
                if (next_node) |child| {
                    node = child;
                    // std.debug.print("found section {s} child_node {any}\n", .{ section, child.type });
                } else {
                    var child_node: ?*Node = null;
                    if (node.placeholder_children.items.len > 1) {
                        // TODO
                        for (node.placeholder_children.items) |child| place: {
                            if (child.max_depth == remaining) {
                                child_node = child;
                                break :place;
                            }
                        }
                    } else if (node.placeholder_children.items.len == 1)
                        child_node = node.placeholder_children.items[0];

                    if (child_node) |n| {
                        if (n.param_name) |name| try params.put(name, section);
                        params_found = true;
                        node = n;
                    } else break;
                }
            }

            // std.debug.print("ended up with node {any}\n", .{node.type});

            if (wildcard_node) |wildcard| {
                node = wildcard;
                try params.put("_", wildcard_param.?);
                params_found = true;
            }

            return self.root_node.handler;
        }

        pub const Node = struct {
            type: Type,
            max_depth: usize, // TODO: what is best here
            parent: ?*Node = null,
            children: std.StringHashMap(*Node),
            handler: HandlerFn,
            param_name: ?[]const u8 = null,
            wildcard_child_node: ?*Node = null,
            placeholder_children: std.ArrayList(*Node),

            pub const Type = enum { normal, wildcard, placeholder };

            pub fn init(allocator: mem.Allocator) !*Node {
                var self = try allocator.create(Node);
                self.type = .normal;
                self.max_depth = 256;
                self.children = std.StringHashMap(*Node).init(allocator);
                self.placeholder_children = std.ArrayList(*Node).init(allocator);
                return self;
            }

            pub fn deinit(self: *Node, allocator: mem.Allocator) void {
                self.placeholder_children.deinit();
                self.children.deinit();
                allocator.destroy(self);
            }
        };
    };
};
